/* ipuz-clue.h
 *
 * Copyright 2022 Jonathan Blandford <jrb@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * SPDX-License-Identifier: (LGPL-2.1-or-later OR MIT)
 */

#pragma once

#include <glib-object.h>
#include <json-glib/json-glib.h>
#include <libipuz/ipuz-enumeration.h>

G_BEGIN_DECLS


typedef enum _IpuzClueDirection
{
  IPUZ_CLUE_DIRECTION_NONE,               /* @ No direction. Used for errors, and to indicate unset IpuzClueId @ */
  IPUZ_CLUE_DIRECTION_ACROSS,             /* @ Across clues @ */
  IPUZ_CLUE_DIRECTION_DOWN,               /* @ Down clues @ */
  IPUZ_CLUE_DIRECTION_DIAGONAL,           /* @ Diagonal clues which go down and to the right @ */
  IPUZ_CLUE_DIRECTION_DIAGONAL_UP,        /* @ Diagonal clues which go up and to the right @ */
  IPUZ_CLUE_DIRECTION_DIAGONAL_DOWN_LEFT, /* @ Diagonal clues which go down and to the left @ */
  IPUZ_CLUE_DIRECTION_DIAGONAL_UP_LEFT,   /* @ Diagonal clues which go up and to the left @ */
  IPUZ_CLUE_DIRECTION_ZONES,              /* @ Clues for explicitly-specified zones @ */
  IPUZ_CLUE_DIRECTION_CLUES,              /* @ Pre-ordered clues (usually alphabetical) @ */
  IPUZ_CLUE_DIRECTION_HIDDEN,             /* @ Clues that aren't shown to the user @ */
  IPUZ_CLUE_DIRECTION_CUSTOM,             /* @ Start of the range of custom directions  @ */
  /* ... */
} IpuzClueDirection;


#define IPUZ_CLUE_DIRECTION_HEADING(dir) (((dir)>=IPUZ_CLUE_DIRECTION_ACROSS)&&((dir)<=IPUZ_CLUE_DIRECTION_DIAGONAL_UP_LEFT))


/* IpuzClueId */

/* Unique id for a clue.
 *
 * Within a puzzle, clues are identified by a pair of:
 * (direction, index)
 * eg.  "the 5th index for the Down clues".
 *
 * If direction == IPUZ_CLUE_DIRECTION_NONE, then the number is
 * meaningless, and this means that there is no clue to be identified
 * (e.g. instead of a NULL pointer for clue). Note, IpuzClueIds aren't
 * stable when editing puzzles.
 */
typedef struct
{
  IpuzClueDirection direction;
  guint index;
} IpuzClueId;

GType       ipuz_clue_id_get_type (void) G_GNUC_CONST;
IpuzClueId *ipuz_clue_id_copy     (const IpuzClueId *clue_id);
void        ipuz_clue_id_free     (IpuzClueId       *clue_id);
gboolean    ipuz_clue_id_equal    (const IpuzClueId *a,
                                   const IpuzClueId *b);

#define IPUZ_CLUE_ID_IS_UNSET(c) ((c)->direction==IPUZ_CLUE_DIRECTION_NONE)
#define IPUZ_TYPE_CLUE_ID (ipuz_clue_id_get_type())


/* Ipuz Cell Coord */


typedef struct
{
  guint row;
  guint column;
} IpuzCellCoord;

GType            ipuz_cell_coord_get_type (void) G_GNUC_CONST;
IpuzCellCoord   *ipuz_cell_coord_copy     (const IpuzCellCoord *coord);
gboolean         ipuz_cell_coord_equal    (const IpuzCellCoord *a,
                                           const IpuzCellCoord *b);

#define IPUZ_TYPE_CELL_COORD (ipuz_cell_coord_get_type())


/* Ipuz Clue */
typedef struct _IpuzClue IpuzClue;
struct _IpuzClue
{
  grefcount ref_count;

  /* Just fields for crosswords, to start. */
  gint number;
  gchar *label;
  gchar *clue_text;
  IpuzClueDirection direction;
  GArray *cells; /* IpuzCellCoord */

  /* < Private > */
  IpuzEnumeration *enumeration;
  IpuzCellCoord location; /* Used for arrowwords */
  gboolean cells_set;     /* TRUE, if cells were loaded vs. calculated */
  gboolean location_set;
};

#define IPUZ_TYPE_CLUE (ipuz_clue_get_type ())


GType              ipuz_clue_get_type              (void);
IpuzClue          *ipuz_clue_new                   (void);
IpuzClue          *ipuz_clue_ref                   (IpuzClue          *clue);
void               ipuz_clue_unref                 (IpuzClue          *clue);
IpuzClue          *ipuz_clue_dup                   (const IpuzClue    *clue);
gboolean           ipuz_clue_equal                 (const IpuzClue    *clue1,
                                                    const IpuzClue    *clue2);
void               ipuz_clue_build                 (IpuzClue          *clue,
                                                    JsonBuilder       *builder);
IpuzClueDirection  ipuz_clue_direction_switch      (IpuzClueDirection  direction);
gint               ipuz_clue_get_number            (IpuzClue          *clue);
void               ipuz_clue_set_number            (IpuzClue          *clue,
                                                    gint               number);
const gchar       *ipuz_clue_get_label             (IpuzClue          *clue);
void               ipuz_clue_set_label             (IpuzClue          *clue,
                                                    const gchar       *label);
const gchar       *ipuz_clue_get_clue_text         (IpuzClue          *clue);
void               ipuz_clue_set_clue_text         (IpuzClue          *clue,
                                                    const gchar       *clue_text);
IpuzEnumeration   *ipuz_clue_get_enumeration       (IpuzClue          *clue);
void               ipuz_clue_set_enumeration       (IpuzClue          *clue,
                                                    IpuzEnumeration   *enumeration);
IpuzClueDirection  ipuz_clue_get_direction         (IpuzClue          *clue);
void               ipuz_clue_set_direction         (IpuzClue          *clue,
                                                    IpuzClueDirection  direction);
IpuzCellCoord      ipuz_clue_get_location          (IpuzClue          *clue,
                                                    gboolean          *location_set);
void               ipuz_clue_set_location          (IpuzClue          *clue,
                                                    IpuzCellCoord      location);
const GArray      *ipuz_clue_get_cells             (IpuzClue          *clue);
void               ipuz_clue_append_cell           (IpuzClue          *clue,
                                                    IpuzCellCoord      coord);
void               ipuz_clue_get_first_cell        (IpuzClue          *clue,
                                                    IpuzCellCoord     *coord);
void               ipuz_clue_get_last_cell         (IpuzClue          *clue,
                                                    IpuzCellCoord     *coord);
gboolean           ipuz_clue_contains_cell         (IpuzClue          *clue,
                                                    IpuzCellCoord      coord);
IpuzClue          *ipuz_clue_new_from_json         (JsonNode          *node);
void               ipuz_clue_ensure_enumeration    (IpuzClue          *clue);


/* Helper functions */
const gchar       *ipuz_clue_direction_to_string   (IpuzClueDirection  direction);
IpuzClueDirection  ipuz_clue_direction_from_string (const gchar       *str);


G_END_DECLS
