/* ipuz-crossword.h
 *
 * Copyright 2022 Jonathan Blandford <jrb@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * SPDX-License-Identifier: (LGPL-2.1-or-later OR MIT)
 */

#pragma once

#include <glib-object.h>
#include <libipuz/ipuz-puzzle.h>
#include <libipuz/ipuz-board.h>
#include <libipuz/ipuz-charset.h>
#include <libipuz/ipuz-clue.h>
#include <libipuz/ipuz-guesses.h>
#include <libipuz/ipuz-symmetry.h>

G_BEGIN_DECLS


#define IPUZ_TYPE_CROSSWORD (ipuz_crossword_get_type ())
G_DECLARE_DERIVABLE_TYPE    (IpuzCrossword, ipuz_crossword, IPUZ, CROSSWORD, IpuzPuzzle);


/**
 * IpuzCrosswordForeachCellFunc:
 * @xword: The iterated #IpuzCrossword
 * @cell: The #IpuzCell
 * @coord: The #IpuzCellCoord of @cell
 * @user_data: data passed to the function
 *
 * The function to be passed to [method@Ipuz.Crossword.foreach_cell].
 *
 * You should not change the size of @xword from within this
 * function. It is safe to change properties of @cell.
 */
typedef void (*IpuzCrosswordForeachCellFunc) (IpuzCrossword       *xword,
                                              IpuzCell            *cell,
                                              const IpuzCellCoord *coord,
                                              gpointer             user_data);


/* Indication of where to physically place the clues on a playing
 * board compared to the grid */
typedef enum
{
  IPUZ_CLUE_PLACEMENT_NULL,   /* @ Auto place clues @ */
  IPUZ_CLUE_PLACEMENT_BEFORE, /* @ Put clues before the puzzle @ */
  IPUZ_CLUE_PLACEMENT_AFTER,  /* @ Put clues after the puzzle @ */
  IPUZ_CLUE_PLACEMENT_BLOCKS, /* @ Put clues in blocks adjacent to entry @ */
} IpuzCluePlacement;

/**
 * IpuzCrosswordForeachClueFunc:
 * @xword: The iterated #IpuzCrossword
 * @direction: The direction of @clue
 * @clue: The next clue in @xword
 * @clue_id: The #IpuzClueId of @clue
 * @user_data: data passed to the function
 *
 * The function to be passed to [method@Ipuz.Crossword.foreach_clue].
 *
 * You should not change the number of clues in @xword from within
 * this function. It is safe to change properties of @clue.
 */
typedef void (*IpuzCrosswordForeachClueFunc) (IpuzCrossword     *xword,
                                              IpuzClueDirection  direction,
                                              IpuzClue          *clue,
                                              IpuzClueId        *clue_id,
                                              gpointer           user_data);

typedef struct _IpuzCrosswordClass IpuzCrosswordClass;
struct _IpuzCrosswordClass
{
  IpuzPuzzleClass parent_class;

  void     (*fix_symmetry)         (IpuzCrossword       *self,
                                    IpuzSymmetry         symmetry,
                                    GArray              *symmetry_coords);
  void     (*fix_numbering)        (IpuzCrossword       *self);
  void     (*fix_clues)            (IpuzCrossword       *self);
  void     (*fix_enumerations)     (IpuzCrossword       *self);
  void     (*fix_styles)           (IpuzCrossword       *self);
  void     (*fix_all)              (IpuzCrossword       *self,
                                    const gchar         *first_attribute_name,
                                    va_list              var_args);
  gboolean (*clue_continues_up)    (IpuzCrossword       *self,
                                    const IpuzCellCoord *coord);
  gboolean (*clue_continues_down)  (IpuzCrossword       *self,
                                    const IpuzCellCoord *coord);
  gboolean (*clue_continues_left)  (IpuzCrossword       *self,
                                    const IpuzCellCoord *coord);
  gboolean (*clue_continues_right) (IpuzCrossword       *self,
                                    const IpuzCellCoord *coord);
  void     (*mirror_cell)          (IpuzCrossword       *self,
                                    const IpuzCellCoord *src_coord,
                                    const IpuzCellCoord *dest_coord,
                                    IpuzSymmetry         symmetry,
                                    IpuzSymmetryOffset   symmetry_offset);
  gboolean (*check_mirror)         (IpuzCrossword       *self,
                                    const IpuzCellCoord *src_coord,
                                    const IpuzCellCoord *target_coord,
                                    IpuzSymmetry         symmetry,
                                    IpuzSymmetryOffset   symmetry_offset);
};


IpuzCrossword     *ipuz_crossword_new                    (void);
guint              ipuz_crossword_get_width              (IpuzCrossword                *self);
guint              ipuz_crossword_get_height             (IpuzCrossword                *self);
void               ipuz_crossword_set_size               (IpuzCrossword                *self,
                                                          gint                          width,
                                                          gint                          height);
IpuzBoard         *ipuz_crossword_get_board              (IpuzCrossword                *self);
gboolean           ipuz_crossword_set_guesses            (IpuzCrossword                *self,
                                                          IpuzGuesses                  *guesses);
IpuzGuesses       *ipuz_crossword_get_guesses            (IpuzCrossword                *self);
IpuzCell          *ipuz_crossword_get_cell               (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *coord);
void               ipuz_crossword_foreach_cell           (IpuzCrossword                *self,
                                                          IpuzCrosswordForeachCellFunc  func,
                                                          gpointer                      user_data);
IpuzCharset       *ipuz_crossword_get_solution_chars     (IpuzCrossword                *self);
IpuzSymmetry       ipuz_crossword_get_symmetry           (IpuzCrossword                *self);

/* Clue sets */
guint              ipuz_crossword_get_n_clue_sets        (IpuzCrossword                *self);
IpuzClueDirection  ipuz_crossword_clue_set_get_dir       (IpuzCrossword                *self,
                                                          guint                         index);
const gchar       *ipuz_crossword_clue_set_get_label     (IpuzCrossword                *self,
                                                          IpuzClueDirection             direction);

/* Clues */
GArray            *ipuz_crossword_get_clues              (IpuzCrossword                *self,
                                                          IpuzClueDirection             direction);
void               ipuz_crossword_foreach_clue           (IpuzCrossword                *self,
                                                          IpuzCrosswordForeachClueFunc  func,
                                                          gpointer                      user_data);
guint              ipuz_crossword_get_n_clues            (IpuzCrossword                *self,
                                                          IpuzClueDirection             direction);
IpuzClue          *ipuz_crossword_get_clue_by_id         (IpuzCrossword                *self,
                                                          IpuzClueId                   *clue_id);
void               ipuz_crossword_unlink_clue            (IpuzCrossword                *self,
                                                          IpuzClue                     *clue);
gboolean           ipuz_crossword_get_id_by_clue         (IpuzCrossword                *self,
                                                          const IpuzClue               *clue,
                                                          IpuzClueId                   *clue_id);
gchar             *ipuz_crossword_get_clue_string_by_id  (IpuzCrossword                *self,
                                                          IpuzClueId                   *clue_id);
gchar             *ipuz_crossword_get_guess_string_by_id (IpuzCrossword                *self,
                                                          IpuzClueId                   *clue_id);
gboolean           ipuz_crossword_clue_guessed           (IpuzCrossword                *self,
                                                          IpuzClue                     *clue,
                                                          gboolean                     *correct);
guint              ipuz_crossword_find_clue              (IpuzCrossword                *self,
                                                          IpuzClue                     *clue);
IpuzClue          *ipuz_crossword_find_clue_by_number    (IpuzCrossword                *self,
                                                          IpuzClueDirection             direction,
                                                          gint                          number);
IpuzClue          *ipuz_crossword_find_clue_by_label     (IpuzCrossword                *self,
                                                          IpuzClueDirection             direction,
                                                          const char                   *label);
IpuzClue          *ipuz_crossword_find_clue_by_coord     (IpuzCrossword                *self,
                                                          IpuzClueDirection             direction,
                                                          const IpuzCellCoord          *coord);

/* Game state */
gboolean           ipuz_crossword_game_won               (IpuzCrossword                *self);

/* Correction API */
void               ipuz_crossword_fix_symmetry           (IpuzCrossword                *self,
                                                          IpuzSymmetry                  symmetry,
                                                          GArray                       *symmetry_coords);
void               ipuz_crossword_fix_numbering          (IpuzCrossword                *self);
void               ipuz_crossword_fix_clues              (IpuzCrossword                *self);
void               ipuz_crossword_fix_enumerations       (IpuzCrossword                *self);
void               ipuz_crossword_fix_styles             (IpuzCrossword                *self);
void               ipuz_crossword_fix_all                (IpuzCrossword                *self,
                                                          const char                   *first_attribute_name,
                                                          ...);
gboolean           ipuz_crossword_clue_continues_up      (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *coord);
gboolean           ipuz_crossword_clue_continues_down    (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *coord);
gboolean           ipuz_crossword_clue_continues_left    (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *coord);
gboolean           ipuz_crossword_clue_continues_right   (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *coord);
void               ipuz_crossword_mirror_cell            (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *src_coord,
                                                          const IpuzCellCoord          *dest_coord,
                                                          IpuzSymmetry                  symmetry,
                                                          IpuzSymmetryOffset            symmetry_offset);
gboolean           ipuz_crossword_check_mirror           (IpuzCrossword                *self,
                                                          const IpuzCellCoord          *src_coord,
                                                          const IpuzCellCoord          *target_coord,
                                                          IpuzSymmetry                  symmetry,
                                                          IpuzSymmetryOffset            symmetry_offset);

/* Public functions */
void               ipuz_crossword_print                  (IpuzCrossword                *self);




G_END_DECLS
