/* ipuz-enumeration.h
 *
 * Copyright 2022 Jonathan Blandford <jrb@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * SPDX-License-Identifier: (LGPL-2.1-or-later OR MIT)
 */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS


#define IPUZ_DELIM_IS_MODIFIER(d)  (d==IPUZ_DELIMINATOR_ALLCAPS||d==IPUZ_DELIMINATOR_CAPITALIZED||d ==IPUZ_DELIMINATOR_FOREIGN)
#define IPUZ_DELIM_IS_SEPARATOR(d) ((guint)d<=IPUZ_DELIMINATOR_APOSTROPHE)

typedef enum _IpuzDeliminator
{
  IPUZ_DELIMINATOR_WORD_BREAK,
  IPUZ_DELIMINATOR_PERIOD,
  IPUZ_DELIMINATOR_DASH,
  IPUZ_DELIMINATOR_APOSTROPHE,

  /* Word Modifiers */
  IPUZ_DELIMINATOR_ALLCAPS,
  IPUZ_DELIMINATOR_CAPITALIZED,
  IPUZ_DELIMINATOR_FOREIGN,
} IpuzDeliminator;

typedef enum _IpuzVerbosity
{
  IPUZ_VERBOSITY_STANDARD, /* "4,4" */
  IPUZ_VERBOSITY_SPARSE,   /* "Two words" */
} IpuzVerbosity;


typedef struct _IpuzEnumeration IpuzEnumeration;


/**
 * IpuzEnumerationForeachDelimFunc:
 * @enumeration: The #IpuzEnumeration to iterate through
 * @delim: The next #IpuzDeliminator being iterated
 * @grid_offset: The offset of @delim within the grid
 * @final_word: Whether the deliminator marks the end of the final word
 * @user_data: data to be passed to the function
 *
 * The function to be passed to [method@Ipuz.Enumeration.foreach_delim].
 *
 * @grid_offset indicates what position within the clue the
 * deliminator can be found. It includes the borders within its count,
 * as well as the cells. For example, consider the enumeration `"^4-4 3"`:
 *
 * `|^| | | - | | | ǁ | | |`
 *
 * This function will be called with %IPUZ_DELIMINATOR_CAPITALIZED at
 * offset 1, %IPUZ_DELIMINATOR_DASH at offset 8, and
 * %IPUZ_DELIMINATOR_WORD_BREAK at offsets 16 and 22. @final_word will
 * be TRUE at offset 22.
 *
 * *Note* this last %IPUZ_DELIMINATOR_WORD_BREAK is implicit, and
 * shouldn't be rendered in the puzzle. It is used to indicates the
 * length of the final word.
 */
typedef void (*IpuzEnumerationForeachDelimFunc) (IpuzEnumeration *enumeration,
                                                 IpuzDeliminator  delim,
                                                 guint            grid_offset,
                                                 gboolean         final_word,
                                                 gpointer         user_data);


#define IPUZ_TYPE_ENUMERATION (ipuz_enumeration_get_type ())


GType            ipuz_enumeration_get_type      (void);
IpuzEnumeration *ipuz_enumeration_new           (const gchar                     *src,
                                                 IpuzVerbosity                    verbosity);
IpuzEnumeration *ipuz_enumeration_ref           (IpuzEnumeration                 *self);
void             ipuz_enumeration_unref         (IpuzEnumeration                 *self);
gboolean         ipuz_enumeration_equal         (const IpuzEnumeration           *enumeration1,
                                                 const IpuzEnumeration           *enumeration2);
gchar           *ipuz_enumeration_get_src       (const IpuzEnumeration           *self);
gchar           *ipuz_enumeration_get_display   (const IpuzEnumeration           *self);
gboolean         ipuz_enumeration_get_has_delim (const IpuzEnumeration           *self);
gint             ipuz_enumeration_delim_length  (const IpuzEnumeration           *self);
void             ipuz_enumeration_foreach_delim (const IpuzEnumeration           *self,
                                                 IpuzEnumerationForeachDelimFunc  func,
                                                 gpointer                         user_data);
gboolean         ipuz_enumeration_valid_char    (gchar                            c);


G_DEFINE_AUTOPTR_CLEANUP_FUNC(IpuzEnumeration, ipuz_enumeration_unref)


G_END_DECLS
